! WHIZARD 1.22 Apr 15 2002
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module partons

  use kinds, only: default !NODEP!
  use diagnostics, only: msg_buffer, msg_message, msg_fatal, msg_bug
  use particles, only: T_QUARK, GLUON, GLUON2, PHOTON, HADRON_REMNANT
  use particles, only: particle_multiplicity, particle_name
  use particles, only: particle, particle_null, particle_new, write
  use density_vectors, only: state_density, create, destroy, write, set
  use beams, only: beam_parameter_set
  use beams, only: strfun_CIRCE_map, strfun_CIRCE
  use beams, only: strfun_ISR
  use beams, only: strfun_EPA
  use pdflib_interface, only: pdflib_init, pdflib_strfun_array !NODEP!

  implicit none
  private

  type, public :: parton_state
     real(kind=default) :: sqrts, rapidity
     integer, dimension(2) :: code
     logical :: polarized_beams, polarized_me
     real(kind=default), dimension(-1:1,-1:1) :: rho_hel
     real(kind=default) :: strfun_factor
     real(kind=default), dimension(2) :: x
     real(kind=default), dimension(-T_QUARK:T_QUARK,2) :: PDF_rho_flv
     integer :: n_col, n_hel, n_flv
     type(state_density), pointer :: rho
     type(particle), dimension(:), pointer :: beam_remnant
     integer :: n_beam_remnants
  end type parton_state


  public :: create, destroy
  public :: write
  public :: structure_functions

  interface create
     module procedure parton_state_create
  end interface
  interface destroy
     module procedure parton_state_destroy
  end interface
  interface write
     module procedure parton_state_write_unit
  end interface

contains

  subroutine parton_state_create &
       & (ps, polarized_beams, polarized_me, sqrts, beam, n_col, n_hel, n_flv)
    type(parton_state), intent(inout) :: ps
    logical, intent(in) :: polarized_beams, polarized_me
    real(kind=default), intent(in) :: sqrts
    type(beam_parameter_set), intent(in), dimension(2) :: beam
    integer, intent(in) :: n_col, n_hel, n_flv
    ps%code = 0
    ps%polarized_beams = polarized_beams
    ps%polarized_me = polarized_me
    ps%sqrts = sqrts
    ps%rapidity = 0
    ps%code = beam%particle_code
    ps%strfun_factor = 0
    ps%x = 0
    ps%rho_hel = 0
    ps%PDF_rho_flv = 0
    ps%n_col = n_col
    ps%n_hel = n_hel
    ps%n_flv = n_flv
    if (n_col > 1)  call msg_fatal &
         & (" Can't happen: Nontrivial initial color distribution.")
    allocate (ps%rho)
    call create (ps%rho, n_col, n_hel, n_flv)
    allocate (ps%beam_remnant(sum(beam%n_strfun)))
    ps%beam_remnant = particle_null
    ps%n_beam_remnants = 0
  end subroutine parton_state_create

  subroutine parton_state_destroy (ps)
    type(parton_state), intent(inout) :: ps
    call destroy (ps%rho)
    deallocate (ps%rho)
  end subroutine parton_state_destroy

  subroutine parton_state_write_unit (u, ps)
    integer, intent(in) :: u
    type(parton_state), intent(in) :: ps
    integer :: j
    write (u,*) "Partonic initial state:"
    write (u,*) "PDG codes        = ", ps%code
    write (u,*) "polarized beams  = ", ps%polarized_beams
    write (u,*) "polarized ME     = ", ps%polarized_me
    write (u,*) "sqrt(s)          = ", ps%sqrts
    write (u,*) "rapidity         = ", ps%rapidity
    write (u,*) "strf.-factor     = ", ps%strfun_factor
    write (u,*) "x(1,2)           = ", ps%x
    write (u,*) "N_col/hel/flv    = ", ps%n_col, ps%n_hel, ps%n_flv
    write (u,*) "PDF flavor probabilities:"
    call write_pdf_rho_flv (1, ps%PDF_rho_flv(:,1))
    call write_pdf_rho_flv (2, ps%PDF_rho_flv(:,2))
    write (u,*) "Helicity probabilities:"
    do j=-1, 1
       write (u,*) ps%rho_hel(:,j)
    end do
    call write (u, ps%rho)
    if (ps%n_beam_remnants > 0) then
       write (u,*) "Beam remnant(s):"
       do j=1, ps%n_beam_remnants
          call write (u, ps%beam_remnant(j))
       end do
    end if
  contains
    subroutine write_pdf_rho_flv (i, rho)
      integer, intent(in) :: i
      real(kind=default), dimension(-T_QUARK:T_QUARK), intent(in) :: rho
      write (u,"(1x,A,I1,A)")  "Beam ", i, ":"
      write (u,11) "d:", rho(1), "u:", rho(2), "s:", rho(3), &
           &       "c:", rho(4), "b:", rho(5), "t:", rho(6)
      write (u,11) "D:", rho(-1), "U:", rho(-2), "S:", rho(-3), &
           &       "C:", rho(-4), "B:", rho(-5), "T:", rho(-6)
      write (u,11) "g:", rho(0)
11    format (6(1x,A,1PE9.3))
    end subroutine write_pdf_rho_flv
  end subroutine parton_state_write_unit

  subroutine structure_functions (ps, beam, r, hel_state_in, flv_state_in)
    type(parton_state), intent(inout) :: ps
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    real(kind=default), dimension(:), intent(in) :: r
    integer, dimension(:,:), intent(in) :: hel_state_in, flv_state_in
    integer :: r_pointer
    r_pointer = 1
    ps%code = beam%particle_code
    ps%strfun_factor = 1
    ps%x = 1
    ps%n_beam_remnants = 0
    ps%beam_remnant = particle_null
    call init_beam_polarization (ps, beam)
    call apply_CIRCE (ps, beam, r, r_pointer)
    call apply_ISR (ps, beam, r, r_pointer)
    call apply_EPA (ps, beam, r, r_pointer)
    call apply_PDF (ps, beam, r, r_pointer)
    call compose_density (ps, beam, hel_state_in, flv_state_in)
  end subroutine structure_functions

  subroutine init_beam_polarization (ps, beam)
    type(parton_state), intent(inout) :: ps
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    real(kind=default), dimension(-1:1,2) :: rho
    integer :: i, j, k
    ps%rho_hel = 0
    if (ps%polarized_me) then
       do i=1, 2
          call beam_spin_density &
               & (rho(-1:1,i), beam(i)%particle_code, beam(i)%polarization)
       end do
       do j=-1, 1
          do k=-1, 1
             ps%rho_hel(j,k) = rho(j,1) * rho(k,2)
          end do
       end do
    else
       ps%rho_hel(0,0) = 1
    end if
  contains
    subroutine beam_spin_density (rho, code, pol)
      real(kind=default), dimension(-1:1), intent(out) :: rho
      integer, intent(in) :: code
      real(kind=default), dimension(:), intent(in) :: pol
      integer :: mult
      real(kind=default), dimension(size(pol)) :: pol_tmp
      rho = 0
      mult = particle_multiplicity (code)
      if (mult>0) then
         pol_tmp = pol(:mult) + (1-sum(pol(:mult)))/mult
      else
         pol_tmp = 0
      end if
      select case(mult)
      case (1)
         rho(0) = 1
      case (2)
         rho(-1) = pol_tmp(1)
         rho(1)  = pol_tmp(2)
      case (3)
         rho(-1:1) = pol_tmp(1:3)
      case default
         write (msg_buffer, "(1x,A,1x,I7,1x,A)") &
              & "Particle:", code, particle_name (code)
         call msg_message
         call msg_message (" Incoming beam: particle spin is undefined.")
         call msg_fatal (" Beam setup failed.")
      end select
    end subroutine beam_spin_density
  end subroutine init_beam_polarization

  subroutine average_beam_polarization (ps, i, mult)
    type(parton_state), intent(inout) :: ps
    integer, intent(in) :: i
    integer, intent(in) :: mult
    integer :: k
    if (ps%polarized_me) then
       select case (i)
       case (1)
          do k=-1, 1
             select case (mult)
             case (1)
                ps%rho_hel (0,k) = sum (ps%rho_hel(:,k))
                ps%rho_hel (-1,k) = 0
                ps%rho_hel (1,k) = 0
             case (2)
                ps%rho_hel (-1,k) = sum (ps%rho_hel(:,k)) / 2
                ps%rho_hel (1,k) = ps%rho_hel (-1,k)
                ps%rho_hel (0,k) = 0
             case (3)
                ps%rho_hel (:,k) = sum (ps%rho_hel(:,k)) / 3
             end select
          end do
       case (2)
          do k=-1, 1
             select case (mult)
             case (1)
                ps%rho_hel (k,0) = sum (ps%rho_hel(k,:))
                ps%rho_hel (k,-1) = 0
                ps%rho_hel (k,1) = 0
             case (2)
                ps%rho_hel (k,-1) = sum (ps%rho_hel(k,:)) / 2
                ps%rho_hel (k,1) = ps%rho_hel (k,-1)
                ps%rho_hel (k,0) = 0
             case (3)
                ps%rho_hel (k,:) = sum (ps%rho_hel(k,:)) / 3
             end select
          end do
       end select
    end if
  end subroutine average_beam_polarization

  subroutine set_beam_remnant (ps, code, x, i)
    type(parton_state), intent(inout) :: ps
    real(kind=default), intent(in) :: x
    integer, intent(in) :: code, i
    integer :: n
    real(kind=default) :: p
    n = ps%n_beam_remnants + 1
    if (n > size(ps%beam_remnant)) &
         & call msg_bug (" Number of allowed beam remnants exceeded")
    p = (1-x) * ps%x(i) * ps%sqrts / 2
    if (i==2)  p = -p
    ps%beam_remnant(n) = particle_new (code, p, 3) 
    ps%n_beam_remnants = n
  end subroutine set_beam_remnant

  subroutine apply_CIRCE (ps, beam, r, r_pointer)
    type(parton_state), intent(inout) :: ps
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    real(kind=default), dimension(:), intent(in) :: r
    integer, intent(inout) :: r_pointer
    real(kind=default), dimension(2) :: x
    integer :: i
    if (all (beam%CIRCE_on)) then
       x = r(r_pointer:r_pointer+1)
       r_pointer = r_pointer + 2
       if (all (beam%CIRCE_map)) then
          call strfun_CIRCE_map (ps%strfun_factor, x(1), x(2), &
               &                 beam(1)%CIRCE_beta, beam(1)%CIRCE_gamma, &
               &                 beam(2)%CIRCE_beta, beam(2)%CIRCE_gamma, &
               &                 beam(1)%CIRCE_prt_out, beam(2)%CIRCE_prt_out)
       else
          call strfun_CIRCE (ps%strfun_factor, x(1), x(2), &
               &                 beam(1)%CIRCE_prt_out, beam(2)%CIRCE_prt_out)
       end if
       do i=1, 2
          if (beam(i)%CIRCE_prt_out == PHOTON) then
             call set_beam_remnant (ps, beam(i)%CIRCE_prt_in, x(i), i)
             call average_beam_polarization &
                  & (ps, i, particle_multiplicity (beam(i)%CIRCE_prt_out))
          else
             call set_beam_remnant (ps, PHOTON, x(i), i)
          end if
       end do
       ps%x = ps%x * x
       ps%code = beam%CIRCE_prt_out
    end if
  end subroutine apply_CIRCE

  subroutine apply_ISR (ps, beam, r, r_pointer)
    type(parton_state), intent(inout) :: ps
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    real(kind=default), dimension(:), intent(in) :: r
    integer, intent(inout) :: r_pointer
    integer :: i
    real(kind=default), dimension(2) :: x
    do i=1, 2
       if (beam(i)%ISR_on) then
          x(i) = r(r_pointer)
          r_pointer = r_pointer + 1
          call strfun_ISR &
               & (ps%strfun_factor, x(i), beam(i)%ISR_power, &
               &  beam(i)%ISR_LLA_order)
          call set_beam_remnant (ps, PHOTON, x(i), i)
          ps%x(i) = ps%x(i) * x(i)
       end if
    end do
  end subroutine apply_ISR

  subroutine apply_EPA (ps, beam, r, r_pointer)
    type(parton_state), intent(inout) :: ps
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    real(kind=default), dimension(:), intent(in) :: r
    integer, intent(inout) :: r_pointer
    real(kind=default), dimension(2) :: x
    integer :: i
    logical :: ok
    do i=1, 2
       if (beam(i)%EPA_on) then
          x(i) = r(r_pointer)
          r_pointer = r_pointer + 1
       end if
    end do
    if (all (beam%EPA_on))  call map_unit_square (x, ps%strfun_factor)
    do i=1, 2
       if (beam(i)%EPA_on) then
          call strfun_EPA (x(i), ps%strfun_factor, beam(i)%EPA_map, &
               &           beam(i)%EPA_x0, beam(i)%EPA_x1, &
               &           beam(i)%EPA_c0, beam(i)%EPA_c1, &
               &           beam(i)%EPA_msq, &
               &           beam(i)%EPA_Q_max**2, beam(i)%EPA_s, &
               &           ok)
          call set_beam_remnant (ps, beam(i)%EPA_prt_in, x(i), i)
          ps%x(i) = ps%x(i) * x(i)
          call average_beam_polarization (ps, i, 2)
          if (.not.ok)  ps%strfun_factor = 0
       end if
    end do
  end subroutine apply_EPA

  subroutine apply_PDF (ps, beam, r, r_pointer)
    type(parton_state), intent(inout) :: ps
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    real(kind=default), dimension(:), intent(in) :: r
    integer, intent(inout) :: r_pointer
    real(kind=default), dimension(2) :: x, scale
    integer :: i
    do i=1, 2
       if (beam(i)%PDF_on) then
          x(i) = r(r_pointer)
          r_pointer = r_pointer + 1
       end if
    end do
    if (all (beam%PDF_on))  call map_unit_square (x, ps%strfun_factor)
    do i=1, 2
       if (beam(i)%PDF_on) then
          if (beam(i)%PDF_running_scale) then
             scale(i) = ps%sqrts * sqrt (x(1)*x(2))
          else
             scale(i) = beam(i)%PDF_scale
          end if
          if (beam(1)%PDF_prt_in /= beam(2)%PDF_prt_in) then
             call pdflib_init &
                  & (beam(i)%PDF_prt_in, beam(i)%PDF_ngroup, &
                  &  beam(i)%PDF_nset, beam(i)%PDF_nfl, beam(i)%PDF_lo, &
                  &  beam(i)%PDF_QCDL4, beam(i)%PDF_mtop)
          end if
          call pdflib_strfun_array (x(i), scale(i), ps%PDF_rho_flv (:,i))
          ps%code(i) = beam(i)%PDF_prt_out
          call set_beam_remnant (ps, HADRON_REMNANT, x(i), i)
          call average_beam_polarization &
               & (ps, i, particle_multiplicity (beam(i)%PDF_prt_out))
          ps%x(i) = ps%x(i) * x(i)
       end if
    end do
  end subroutine apply_PDF

  subroutine map_unit_square (x, factor, power)
    real(kind=default), dimension(2), intent(inout) :: x
    real(kind=default), intent(inout) :: factor
    real(kind=default), intent(in), optional :: power
    real(kind=default) :: xx, yy
    xx = x(1)
    yy = x(2)
    if (present(power)) then
       if (x(1) > 0 .and. power > 1) then
          xx = x(1)**power
          factor = factor * power * xx / x(1)
       end if
    end if
    if (xx /= 0) then
       x(1) = xx ** yy
       x(2) = xx / x(1)
       factor = factor * abs (log (xx))
    else
       x = 0
    end if
  end subroutine map_unit_square

  subroutine compose_density (ps, beam, hel_state_in, flv_state_in)
    type(parton_state), intent(inout) :: ps
    type(beam_parameter_set), dimension(2), intent(in) :: beam
    integer, dimension(:,:), intent(in) :: hel_state_in, flv_state_in
    integer :: i_hel, i_flv, i
    integer, dimension(2) :: hel, flv
    real(kind=default) :: val, factor
    do i_flv = 1, ps%n_flv
       flv = flv_state_in (:,i_flv)
       factor = ps%strfun_factor
       do i=1, 2
          if (beam(i)%PDF_on) then
             select case (flv(i))
             case (GLUON, GLUON2)
                flv(i) = 0
             end select
             factor = factor * ps%PDF_rho_flv(flv(i),i)
          end if
       end do
       if (ps%polarized_me) then
          do i_hel = 1, ps%n_hel
             hel = hel_state_in (:,i_hel)
             val = factor * ps%rho_hel(hel(1),hel(2))
             call set (ps%rho, 1, i_hel, i_flv, val)
          end do
       else
          val = factor
          call set (ps%rho, 1, 1, i_flv, val)
       end if
    end do
  end subroutine compose_density


end module partons

